<?php
/**
 * Created by JetBrains PhpStorm.
 * User: lucassouza
 * Date: 01/11/11
 * Time: 09:19
 * To change this template use File | Settings | File Templates.
 */
namespace Setforms2\Runtime;

use \Setforms2\Runtime\I18n\Replacer;
use \Setforms2\Runtime\Interfaces\IRequest;
use \Setforms2\Runtime\Interfaces\IResponse;
use \Setforms2\Shared\Connection\Exception\LimitRowsExceededException;
use \Setforms2\Runtime\Dto\Method as MethodToCall;
use \Setforms2\Shared\Util\Util;
use \Setforms2\Runtime\Interfaces\IUserInfo;

class Framework
{

    /** Constant to identify a abort command. Mostly used in validation events like "beforeSave" */
    const ABORT = 'FRAMEWORK_ABORT';

    private $messages;
    private $fieldsToUpdate;
    private $data;
    private $dataset;
    private $args;
    private $fields;
    private $windowToOpen;
    /** @var Session */
    private $session;
    private $selectData;
    private $methodsToCall;
    private $errors;
    private $currentMapper;
    private $aborted;
    private $extras = array();
    private $success = array();
    private $wizardRecord = array();
    private $filesToDownload = array();
    /**
     * @var \Setforms2\Runtime\Controller
     */
    private $controller;

    /**
     * @var Framework
     */
    private static $instance;

    private function __construct()
    {

        $this->aborted = false;
        //error_reporting(E_ALL|E_STRICT);
        set_error_handler(array($this, 'errorHandler'));
        set_exception_handler(array($this, 'exceptionHandler'));
        register_shutdown_function(array($this, 'fatalErrorHandler'));
        define('ABORT', self::ABORT);
    }

    public function errorHandler($errno, $errstr, $errfile, $errline)
    {
        if (! (error_reporting() & $errno)) {
            // This error code is not included in error_reporting
            return;
        }

        $msg = '';
        switch ($errno) {
            case E_USER_ERROR:
                $msg .= "[ERROR] [$errno] $errstr\n";
                break;
            case E_USER_WARNING:
                $msg .= "[WARNING] [$errno] $errstr\n";
                break;
            case E_USER_NOTICE:
                $msg .= "[NOTICE] [$errno] $errstr\n";
                break;
            default:
                $msg .= "Unknown error type: [$errno] $errstr\n";
                break;
        }

        //@todo implement warning levels that will define if file and line, or any other information, should be showed.
        tk_message("{$msg} at {$errfile}({$errline}).");
        /* Don't execute PHP internal error handler */
        return true;
    }

    public function exceptionHandler(\Exception $exception)
    {
        ob_end_clean();
        header('Content-type: application/json');
        if ($this->controller) {
            $this->controller->buildResponse();
            $response = $this->controller->getResponse();
        } else {
            $response = new \Setforms2\Runtime\Response();
        }
        if ($exception instanceof \Setforms2\Runtime\Exception\AbortException) {
            if($exception->isShowMessage()) {
                if($exception->isShowMessageOnly()) {
                    $response->addMessage($exception->getMessage(), $exception->getMessageType());
                } else {
                    $response->setErrors(array($exception->getMessage()));
                }

            }
        } else {
            $response->setErrors(
                array($exception->getMessage().
                    "<br>
                    <details>
                    <summary> Server Error Detail </summary>
                    <pre>{$exception->getTraceAsString()}</pre>
                    </details>")
            );
        }
        $response->setExtras(Array("aborted" => true));
        echo $response->toJson();
        die();
    }

    public function fatalErrorHandler()
    {
        if (@is_array($e = @error_get_last())) {
            $code = isset($e['type']) ? $e['type'] : 0;
            $file = isset($e['file']) ? $e['file'] : '';
            $line = isset($e['line']) ? $e['line'] : '';
            $msg = "[FATAL ERROR] ";
            //@todo implement warning levels that will define if file and line, or any other information, should be showed.
            $msg .= isset($e['message']) ? $e['message'] : '';
            $msg .= " at {$file}({$line}).";

            $this->logFatal($msg);

            if ($code > 0) {
                ob_end_clean();
                header('Content-type: application/json');
                $response = new \Setforms2\Runtime\Response();
                $response->setErrors(array($msg));
                $response->setSuccess(false);

                echo $response->toJson();
            }
        }

    }

    public function logFatal($msg){
        $dir = APP_PATH.'/log/';
        if (!realpath($dir)){
            mkdir($dir);
        }
        $d = date('d');
        $m = date('m');
        $y = date('Y');
        $h = date('h');
        $file = $dir.$y.$m.$d;
        file_put_contents($file, "\n".$h.' '.$msg, FILE_APPEND | LOCK_EX);
    }

    public function handleDbError(\Exception $exception) {
        static $dbDriverExceptions = array (
            'Doctrine2\DBAL\Driver\OCI8\OCI8Exception'
        );

        if (in_array(get_class($exception->getPrevious()), $dbDriverExceptions)) {
            $dbErrorHandler = \Setforms2\Shared\Config::getInstance()->getProjectDbErrorHandler();
            return call_user_func(array($dbErrorHandler, 'handleMessage'), $exception->getPrevious());
        }

        return false;
    }

    /**
     *
     * @return Framework
     */
    public static function getInstance()
    {
        if (static::$instance === null) {
            static::$instance = new self;
        }

        return static::$instance;
    }

    public function init()
    {
        $request = Request::fromGlobals();
        $this->validateRevisionId($request);
        $this->initRequest($request);
        $this->initArgs($request);
        $response = new Response();
        $this->controller = new Controller($request, $response);
        $this->controller->dispatch();

        header('Content-type: application/json');
        echo $response->toJson();
    }

    public function initRequest(IRequest $request)
    {
        $dataset = $request->getDataset();
        if ($dataset) {
            $this->setDataset($dataset);
            $this->initFields($dataset);
        }
    }

    public function initFields($dataset)
    {
        $this->fields = array();
        foreach ($dataset as $row) {
            if(count($row) === 0 || (isset($row['__group']) && $row['__group'])) continue;

            $id = $row['_ID'];
            foreach ($row as $name => $value) {
                $this->fields[$name][$id] = $value;
            }
        }
    }

    public function prepareResponse(IResponse $response)
    {
        $response->setData($this->getData());
        $response->setFieldsToUpdate($this->getFieldsToUpdate());
        $response->setSelectData($this->getSelectData());
        $methodsToCall = $this->getMethodsToCallAsArray();
        $response->setMethodsToCall($methodsToCall);
        $response->setExtras($this->getExtras());

        $this->prepareWindowToOpen();
        $response->setWindowToOpen(Util::objectToArray($this->getWindowToOpen()));
        $response->setErrors($this->getErrors());
        $response->setSuccess($this->hadSuccess());
        $response->setSuccessById($this->getSuccess());

        $response->setMessages($this->getMessages());
        $response->setFilesToDownload($this->getFilesToDownload());
        return $response;
    }

    private function getMethodsToCallAsArray()
    {
        $methodsToCall = $this->getMethodsToCall();
        if (is_array($methodsToCall)) {
            $methodsToCall = array_map(array('\Setforms2\Shared\Util\Util', 'objectToArray'), $methodsToCall);
        }

        return $methodsToCall;
    }

    public function getMessages()
    {
        return $this->messages;
    }

    public function setMessages($messages)
    {
        $this->messages = $messages;
    }

    public function addMessage($message, $type = 'info', $secsToExpire = null, $isAlert = false)
    {
        $this->messages[$type][] = array(
            'message' => $message,
            'expire' => $secsToExpire,
            'isAlert' => $isAlert
        );
    }

    public function getFieldsToUpdate()
    {
        return $this->fieldsToUpdate;
    }

    public function getFieldsToUpdateByTab($tab_id = null)
    {
        if ($tab_id == null) {
            $tab_id = self::getTabId();
        }

        return isset($this->fieldsToUpdate[$tab_id]) ? $this->fieldsToUpdate[$tab_id] : array();
    }

    public function setFieldsToUpdate($fieldsToUpdate)
    {
        $this->fieldsToUpdate = $fieldsToUpdate;
    }

    public function addFieldToUpdate($name, $value)
    {
        $this->fieldsToUpdate[$name] = $value;
    }

    public function addFileToDownload(\TkFile $file)
    {
        \Setforms2\Runtime\File\Repository::getInstance()->add($file);
        $this->filesToDownload[] = $file->getId();
    }
    
    public function getFilesToDownload() {
        return $this->filesToDownload;
    }

    public function getData()
    {
        return $this->data;
    }

    public function addData($data, $tabName = null, $windowName = null)
    {
        $tabId = self::getTabId($tabName, $windowName);
        $this->data[$tabId] = $data;
    }

    public function setData($data)
    {
        $this->data = $data;
    }

    public function fetchAll($sql, $params = array())
    {
        $conn = Connection::getInstance();
		return $conn->fetchAll($sql, $params);
    }

    public function prepareData($data)
    {
        if ($data) {
            $data = $this->addIdToRows($data);
        }

        $this->setData($data);
    }

    /**
     * @autor jorgefonseca
     * @todo Paulo retirou esse metodo, verificar porque.
     */
    public function catchLimitRowsExceededException()
    {
        tk_show_message_by_name('TK_LIMIT_ROWS_EXCEEDED', array(), 'warning');
    }

    public function addIdToRows($data, $index = 0)
    {
        $data = array_map(function($item) use (&$index)
        {
            $item['_ID'] = ++$index;
            $item['_STATE'] = '';
            $item['_CHECKED'] = false;
            return $item;
        }, $data);
        return $data;
    }

    public function getDataset()
    {
        return $this->dataset;
    }

    public function setDataset($dataset)
    {
        $this->dataset = $dataset;
    }

    public function getFields()
    {
        return $this->fields;
    }

    public function setFields($fields)
    {
        $this->fields = $fields;
    }

    public function getFieldValue($name)
    {
        return isset($this->fields[$name]) ? $this->fields[$name] : null;
    }

    public function setFieldValue($name, $value, $rowId, $tab_name = null, $window_name = null)
    {
        $tab_id = self::getTabId($tab_name, $window_name);
        $this->fieldsToUpdate[$tab_id][$rowId][$name] = $value;
    }

    public function addFieldValue($name, $value)
    {
        $name = strtolower($name);
        $this->fieldsToUpdate[$name][] = $value;
    }

    public function openWindow($window)
    {
        $this->setWindowToOpen($window);
    }

    public function resetFieldsValues()
    {
        $fields = $this->getFields();
        $this->fields = array();
        foreach ($fields as $name => $value){
            $this->fields[$name][0] = null;
        }
    }

    /**
     * @return \Setforms2\Runtime\Dto\WindowToOpen
     */
    public function getWindowToOpen()
    {
        return $this->windowToOpen;
    }

    public function setWindowToOpen($windowToOpen)
    {
        $this->windowToOpen = $windowToOpen;
    }

    public function isLoginWindowRequest()
    {
        if (isset($this->args[1])) {
            return $this->args[1] === 'LOGIN_WINDOW_REQUEST';
        }

        return false;
    }

    private function prepareWindowToOpen()
    {
        $windowToOpen = $this->getWindowToOpen();
        if ($windowToOpen) {
            $html_path = $this->getWindowFileName($windowToOpen);
            $json_path = $this->getWindowJSONPath($windowToOpen);
            if (file_exists($html_path) && file_exists($json_path)) {
                $window_content = file_get_contents($html_path);
                $window_json = file_get_contents($json_path);
                $window_json = Util::replaceTokenJson(json_decode($window_json, true));
                $window_content = $this->translateWindow($window_content, $windowToOpen->getName());
                if (!$this->isLoginWindowRequest()) {
                    $window_content = Util::replaceToken(stripcslashes($window_content));
                    $window_content = \Setforms2\Runtime\AccessControl\WindowAccessControl::processDocument($window_content);
                }

                $windowToOpen->setContent($window_content);
                $windowToOpen->setFrame(json_decode(json_encode($window_json)));
            }
        }
    }

    private function applyTranslateByMapper($mapperName, $windowContent) {
        try {
            $replacer = Replacer::getFor($mapperName);
            $windowContent = $replacer->replace($windowContent);
        } catch (\Setforms2\Runtime\I18n\Exception\I18nMapperNotFoundException $e) {
            //nothing to do, the window just doesn't have that elements to translate, "eg. a window without field groups".
        }
        return $windowContent;
    }

    private function translateWindow($windowContent, $windowName)
    {
        $windowName = "\\Windows\\".\SetformsInflector::classify($windowName)."\\I18n\\";
        $windowContent = $this->applyTranslateByMapper("\\I18n\\Messages", $windowContent);
        $windowContent = $this->applyTranslateByMapper($windowName."Windows", $windowContent);
        $windowContent = $this->applyTranslateByMapper($windowName."Tabs", $windowContent);
        $windowContent = $this->applyTranslateByMapper($windowName."Fields", $windowContent);
        $windowContent = $this->applyTranslateByMapper($windowName."Actions", $windowContent);
        $windowContent = $this->applyTranslateByMapper($windowName."Panels", $windowContent);
        $windowContent = $this->applyTranslateByMapper($windowName."Fieldgroups", $windowContent);
        return $windowContent;
    }

    private function getWindowFileName(Dto\WindowToOpen $windowToOpen)
    {
        $file_dir = $this->getWindowPath($windowToOpen);
        $windowName = $windowToOpen->getName();
        $file_path = "$file_dir/{$windowName}.html";
        return $file_path;
    }

    private function getWindowJSONPath(Dto\WindowToOpen $windowToOpen)
    {
        $file_dir = $this->getWindowPath($windowToOpen);
        $windowName = $windowToOpen->getName();
        $file_path = "$file_dir/{$windowName}.json";
        return $file_path;
    }

    private function getWindowPath(Dto\WindowToOpen $windowToOpen)
    {
        $window_path = realpath(APP_PATH.\Setforms2\Shared\Config::getInstance()->getProjectGeneratedPath());
        $windowName = $windowToOpen->getName();
        $directoryName = \SetformsInflector::classify($windowName);
        $file_path = "{$window_path}/Windows/{$directoryName}";
        return $file_path;
    }

    public function execute($sql, $params = array())
    {
        $conn = Connection::getInstance();
        $result = $conn->execute($sql, $params);
        $result = $result->rowCount();
        return $result === 0 ? tk_abort(tk_get_message_by_name('TK_RESULT_EMPTY')) : $result;
    }

    public function getArgs()
    {
        return $this->args;
    }

    private function setArgs($args)
    {
        $this->args = $args;
    }

    private function initArgs($request)
    {
        $this->setArgs($request->getArgs());
    }

    public function getSession()
    {
        if ($this->session == null) {
            $this->session = new Session();
        }

        return $this->session;
    }

    public function getSelectData()
    {
        return $this->selectData;
    }

    public function setSelectData($field_name, $select_data, $tab_name = null, $window_name = null)
    {
        $tab_id = self::getTabId($tab_name, $window_name);
        $this->selectData[$tab_id][$field_name] = $select_data;
    }

    public function getMethodsToCall()
    {
        return $this->methodsToCall;
    }

    public function setMethodsToCall($methodsToCall)
    {
        $this->methodsToCall = $methodsToCall;
    }

    public function addMethodToCall($methodToCall)
    {
        $this->methodsToCall[] = $methodToCall;
    }

    public function getErrors()
    {
        return $this->errors;
    }

    public function setErrors($errors)
    {
        $this->errors = $errors;
    }

    public function addError($id, $error)
    {
        $this->errors[$id] = $error;
        $this->addSuccess($id, false);
    }

    public function getResultsetLimit()
    {
        $limit = 99999999999;
        if ($current_mapper = $this->getCurrentMapper()) {
            $limit = $current_mapper->getResultsetLimit();
        }

        return $limit;
    }

    /**
     * @return \Setforms2\Runtime\Mapper\Crud|\Setforms2\Runtime\Mapper\Base|\Setforms2\Runtime\Mapper\Tree|\Setforms2\Runtime\Mapper\Mixed
     */
    public function getCurrentMapper()
    {
        return $this->currentMapper;
    }

    public function setCurrentMapper($currentMapper)
    {
        $this->currentMapper = $currentMapper;
    }

    public function isAbort($value)
    {
        if ($value === self::ABORT) {
            $this->aborted = true;
            return true;
        }

        return false;
    }

    public function setUserInfo(IUserInfo $userInfo)
    {
        $this->getSession()->set('userinfo_object', $userInfo);
    }

    /**
     * @return IUserInfo|null
     */
    public function getUserInfo()
    {
        $user_info = $this->getSession()->get('userinfo_object');
        if (!$user_info) {
            $user_info = new \Setforms2\Runtime\Dto\UserInfo();
            $this->setUserInfo($user_info);
        }

        return $user_info;
    }

    public function getExtras()
    {
        return $this->extras;
    }

    public function setExtras($extras)
    {
        $this->extras = $extras;
    }

    public function addExtra($key, $value, $tabName = null, $windowName = null)
    {
        $tabId = self::getTabId($tabName, $windowName);
        $this->extras[$tabId][$key] = $value;
    }

    public function getSuccess()
    {
        return $this->success;
    }

    public function setSuccess($success)
    {
        $this->success = $success;
    }

    public function addSuccess($id, $success)
    {
        $this->success[$id] = $success;
    }

    public function hadSuccess()
    {
        $success = $this->getSuccess();
        foreach ($success as $succeeded) {
            if ($succeeded) return true;
        }

        return empty($this->errors) && ! $this->aborted;
    }

    public static function getTabId($tab_name = null, $window_name = null)
    {
        $currentMapper = Framework::getInstance()->getCurrentMapper();
        if ($currentMapper instanceof \Setforms2\Runtime\Mapper\Window) {
            throw new \Exception('Window has no tab yet.');
        }
        if ($tab_name === null) {
            $tab_name = $currentMapper->getTabName();
        }

        if ($window_name === null) {
            $window_name = $currentMapper->getWindowName();
        }

        $tab_id = strtolower($window_name.":".$tab_name);

        /* If has menu params */
        if($currentMapper != null) $menuParams = $currentMapper->getMenuParams();

        if (empty($menuParams) === false) {
            $menu_name = $currentMapper->getMenuName();
            $tab_id = strtolower($menu_name).'_'.$tab_id;
        }

        return $tab_id;
    }

    public static function getMapper($window_name = null, $tab_name = null) {
        $mapper = self::getInstance()->getCurrentMapper();
        if ($window_name && $tab_name && !($mapper->getWindowName() == $window_name && $mapper->getTabName() == $tab_name)) {
            $windowNamespace = \SetformsInflector::classify($window_name);
            $tabNamespace = \SetformsInflector::classify($tab_name);
            $custom_name = "\\Windows\\{$windowNamespace}\\{$tabNamespace}\\Custom";
            $mapper_name = "\\Windows\\{$windowNamespace}\\{$tabNamespace}\\Mapper";
            $mapper_class_name = (class_exists($custom_name)) ? $custom_name : $mapper_name;

            $mapper = new $mapper_class_name();
        }

        return $mapper;
    }

    public function getWizardRecord($key) {
        if(isset($this->wizardRecord[$key])) {
            return $this->wizardRecord[$key];
        }

        return null;
    }
    
    public function addWizardRecord($dataset) {
        $this->wizardRecord = array_merge($this->wizardRecord, $dataset);
    }

    public function getProjectMapper() {
        $project_name = \Setforms2\Shared\Config::getInstance()->getProjectName();
        $mapper_name = \SetformsInflector::classify($project_name).'Mapper';
        $m_name = "\\Project\\".$mapper_name;
        return new $m_name();
    }

    /**
     * @param $request
     */
    protected function validateRevisionId($request) {
        $filename = realpath(APP_PATH . '/public/build.json');
        $json = file_get_contents($filename);
        $data = json_decode($json);
        $revision = $data->revision;
        if (($revision != $request->getRevisionId() && !defined('TK_IGNORE_REVISION_ID'))) {
            $message = tk_get_message_by_name('TK_SERVER_UPDATED');
            tk_warning($message);
        }
    }

}